@@ -5,16 +5,23 @@
    
    This file is public domain.  */
 
-/* $Id: jmpextra.c,v 1.11 2000/03/20 02:55:50 hideyuki Exp $ */
-
 #define	EXTERN /* Instantiate data from {tex,mf,mp}d.h here.  */
 
 /* This file is used to create texextra.c etc., with this line
    changed to include texd.h, mfd.h, or mpd.h.  The ?d.h file is what
    #defines TeX or MF or MP, which avoids the need for a special
    Makefile rule.  */
+#ifdef WIN32
+#include <windows.h>
+#include <fcntl.h>
+#include <process.h>
+int osfhandle;
+char prockanjicode = 2; /* 0: jis, 1: euc, 2: sjis */
+#ifndef WIN32WIN
+#define WIN32WIN 1
+#endif
+#endif
 #include "mpd.h"
-#include "jmpost.h"
 
 #include <kpathsea/c-ctype.h>
 #include <kpathsea/line.h>
@@ -83,7 +90,7 @@ extern struct tm *localtime ();
 #ifdef MP
 const_string local_maintainer="Hideyuki Suzuki";
 const_string this_bug_address="hideyuki@sat.t.u-tokyo.ac.jp";
-#define BANNER "This is jMetaPost, Version 0.04, based on MetaPost, Version 0.641"
+#define BANNER "This is jMetaPost, Version 0.641-0.04"
 #define COPYRIGHT_HOLDER "AT&T Bell Laboratories"
 #define AUTHOR "John Hobby"
 #define PROGRAM_HELP JMPHELP
@@ -109,25 +116,53 @@ static string user_progname;
 /* The C version of what might wind up in DUMP_VAR.  */
 static string dump_name;
 
+/* The C version of the jobname, if given. */
+static const_string job_name;
+
+/* Full source file name. */
+extern string fullnameoffile;
+
 /* The filename for dynamic character translation, or NULL.  */
 string translate_filename;
+string default_translate_filename;
+
+/* Needed for --src-specials option. */
+static char *last_source_name;
+static int last_lineno;
+static boolean srcspecialsoption = false;
+static void parse_src_specials_option P1H(const_string);
 
 /* The main body of the WEB is transformed into this procedure.  */
 extern TEXDLL void mainbody P1H(void);
 
-static void maybe_parse_first_line P1H(void);
+/* Parsing a first %&-line in the input file. */
+static void parse_first_line P1H(void);
+
 static void parse_options P2H(int, string *);
 
+#ifdef MP
+/* name of TeX program to pass to makempx */
+static string mpost_tex_program = "";
+#endif
+
+#ifdef __STDC__
+/*
+#ifdef WIN32
+string locale_name = ".ACP";
+#else
+*/
+string locale_name = "";
+/*
+#endif
+*/
+#endif
+
 /* The entry point: set up for reading the command line, which will
    happen in `topenin', then call the main body.  */
 
 void TEXDLL
 maininit P2C(int, ac, string *, av)
 {
-#if defined(__STDC__)
-  /* "" means: get value from env. var LC_ALL, LC_CTYPE, or LANG */
-  setlocale(LC_CTYPE, "");
-#endif
 
   /* Save to pass along to topenin.  */
   argc = ac;
@@ -141,18 +176,35 @@ maininit P2C(int, ac, string *, av)
      the web (which would read the base file, etc.).  */
   parse_options (ac, av);
   
+#if defined(__STDC__)
+  /* Need to delay it because of win32 `-oem' option.  Default value ""
+     means: get value from env. var LC_ALL, LC_CTYPE, or LANG */
+  setlocale(LC_CTYPE, locale_name);
+#endif
+
   /* Do this early so we can inspect program_invocation_name and
      kpse_program_name below, and because we have to do this before
      any path searching.  */
   kpse_set_program_name (argv[0], user_progname);
 
   /* If no dump default yet, and we're not doing anything special on
-     this run, look at the first line of the main input file for a
-     %&<dumpname> specifier.  */
-  if (!dump_name || !translate_filename) {
-    maybe_parse_first_line ();
+     this run, we may want to look at the first line of the main input
+     file for a %&<dumpname> specifier.  */
+  if (!parsefirstlinep) {
+    string parse_first_line = kpse_var_value ("parse_first_line");
+    parsefirstlinep = (parse_first_line
+                       && (*parse_first_line == 't'
+                           || *parse_first_line == 'y'
+                           || *parse_first_line == '1'));
+  }
+  if (parsefirstlinep && (!dump_name || !translate_filename)) {
+    parse_first_line ();
+  }
+  /* Check whether there still is no translate_filename known.  If so,
+     use the default_translate_filename. */
+  if (!translate_filename) {
+    translate_filename = default_translate_filename;
   }
-  
   /* If we're preloaded, I guess everything is set up.  I don't really
      know any more, it's been so long since anyone preloaded.  */
   if (readyalready != 314159) {
@@ -200,7 +252,13 @@ maininit P2C(int, ac, string *, av)
 #ifdef MF
   kpse_set_program_enabled (kpse_mf_format, MAKE_TEX_MF_BY_DEFAULT,
                             kpse_src_compile);
+  kpse_set_program_enabled (kpse_base_format, MAKE_TEX_FMT_BY_DEFAULT,
+                            kpse_src_compile);
 #endif /* MF */
+#ifdef MP
+  kpse_set_program_enabled (kpse_mem_format, MAKE_TEX_FMT_BY_DEFAULT,
+                            kpse_src_compile);
+#endif /* MP */
 #ifdef TeX
 #ifdef Omega
   kpse_set_program_enabled (kpse_ocp_format, MAKE_OMEGA_OCP_BY_DEFAULT,
@@ -208,14 +266,15 @@ maininit P2C(int, ac, string *, av)
   kpse_set_program_enabled (kpse_ofm_format, MAKE_OMEGA_OFM_BY_DEFAULT,
                             kpse_src_compile);
   kpse_set_program_enabled (kpse_tfm_format, false, kpse_src_compile);
-  kpse_set_program_enabled (kpse_tex_format, MAKE_TEX_TEX_BY_DEFAULT,
-                            kpse_src_compile);
 #else
   kpse_set_program_enabled (kpse_tfm_format, MAKE_TEX_TFM_BY_DEFAULT,
                             kpse_src_compile);
+#endif /* !Omega */
   kpse_set_program_enabled (kpse_tex_format, MAKE_TEX_TEX_BY_DEFAULT,
                             kpse_src_compile);
-#endif /* !Omega */
+  kpse_set_program_enabled (kpse_fmt_format, MAKE_TEX_FMT_BY_DEFAULT,
+                            kpse_src_compile);
+
   if (!shellenabledp) {
     string shell_escape = kpse_var_value ("shell_escape");
     shellenabledp = (shell_escape
@@ -229,7 +288,6 @@ maininit P2C(int, ac, string *, av)
 #endif /* TeX */
 }
 
-#ifndef WIN32
 /* The entry point: set up for reading the command line, which will
    happen in `topenin', then call the main body.  */
 
@@ -240,14 +298,32 @@ main P2C(int, ac,  string *, av)
   _wildcard (&ac, &av);
   _response (&ac, &av);
 #endif
+#ifdef WIN32 /* A. Kakuto */
+  int ai;
+  char *ap;
+  if(ac > 1) {
+    for(ai = 1; ai < ac; ai++) {
+      if(ap = strrchr(av[ai], '.')) {
+	while(*ap) {
+	  *ap = tolower(*ap);
+	  ap++;
+	}
+      }
+    }
+  }
+  getlongname(av[0]);
+  osfhandle = (int)_get_osfhandle(fileno(stdout));
+#endif /* WIN32 */
 
   maininit(ac, av);
 
   /* Call the real main program.  */
   mainbody ();
+#ifdef WIN32
+  win32fflush(stdout);
+#endif
   return EXIT_SUCCESS;
 } 
-#endif /* ! WIN32 */
 
 /* This is supposed to ``open the terminal for input'', but what we
    really do is copy command line arguments into TeX's or Metafont's
@@ -301,8 +377,10 @@ topenin P1H(void)
    incrementally.  Shamim Mohamed adapted it for Web2c.  */
 #if defined (TeX) && defined (IPC)
 
+#ifndef WIN32
 #include <sys/socket.h>
 #include <fcntl.h>
+#endif
 #ifndef O_NONBLOCK /* POSIX */
 #ifdef O_NDELAY    /* BSD */
 #define O_NONBLOCK O_NDELAY
@@ -310,35 +388,61 @@ topenin P1H(void)
 #ifdef FNDELAY     /* NeXT */
 #define O_NONBLOCK O_FNDELAY
 #else
-what the fcntl? cannot implement IPC without equivalent for O_NONBLOCK.
+#ifdef WIN32
+#ifdef FIONBIO
+#define O_NONBLOCK FIONBIO
+#endif
+#endif /* WIN32 */
+/* what the fcntl? cannot implement IPC without equivalent for O_NONBLOCK. */
 #endif /* no FNDELAY */
 #endif /* no O_NDELAY */
 #endif /* no O_NONBLOCK */
 
 #ifndef IPC_PIPE_NAME /* $HOME is prepended to this.  */
+#ifndef WIN32
 #define IPC_PIPE_NAME "/.TeXview_Pipe"
+#else
+#define IPC_LOCAL_HOST "127.0.0.1"
+#define FIXED_PORT     (unsigned short)4242
+#endif
 #endif
 #ifndef IPC_SERVER_CMD /* Command to run to start the server.  */
+#ifdef WIN32
+#define IPC_SERVER_CMD "texview.exe"
+#else
 #define IPC_SERVER_CMD "open `which TeXview`"
 #endif
+#endif
 
 struct msg
 {
-  short namelength; /* length of auxiliary data */
-  int eof;   /* new eof for dvi file */
+  int   namelength; /* length of auxiliary data */
+  int   eof;        /* new eof for dvi file */
 #if 0  /* see usage of struct msg below */
   char more_data[0]; /* where the rest of the stuff goes */ 
 #endif
 };
 
+#ifndef WIN32
 static char *ipc_name;
 static struct sockaddr *ipc_addr;
-static int ipc_addr_len;
+#else
+static struct sockaddr_in ipc_addr;
+static unsigned long remote_addr;
+#endif
+
+static int ipc_addr_len = 0;
 
 static int
 ipc_make_name P1H(void)
 {
   if (ipc_addr_len == 0) {
+#ifdef WIN32
+    ipc_addr.sin_family = AF_INET;
+    ipc_addr.sin_addr.s_addr = remote_addr;
+    ipc_addr.sin_port = htons(FIXED_PORT);
+    ipc_addr_len = sizeof(struct sockaddr_in);
+#else
     string s = getenv ("HOME");
     if (s) {
       ipc_addr = xmalloc (strlen (s) + 40);
@@ -348,49 +452,110 @@ ipc_make_name P1H(void)
       strcat (ipc_name, IPC_PIPE_NAME);
       ipc_addr_len = strlen (ipc_name) + 3;
     }
+#endif
   }
   return ipc_addr_len;
 }
 
-
+#ifdef WIN32
+static int sock = INVALID_SOCKET;
+unsigned long  sockcntl;
+#else
 static int sock = -1;
+#endif
 
 static int
 ipc_is_open P1H(void)
 {
+#ifdef WIN32
+   return sock != INVALID_SOCKET;
+#else
    return sock >= 0;
+#endif
 }
 
-
 static void
 ipc_open_out P1H(void) {
+#ifdef WIN32
+  struct WSAData wsaData;
+  int nCode;
+#endif
 #ifdef IPC_DEBUG
   fputs ("tex: Opening socket for IPC output ...\n", stderr);
 #endif
+#ifdef WIN32
+  if (sock != INVALID_SOCKET) {
+    return;
+  }
+  if ((nCode = WSAStartup(MAKEWORD(1, 1), &wsaData)) != 0) {
+    fprintf(stderr,"WSAStartup() returned error code %d.\n", nCode);
+    sock = INVALID_SOCKET;
+    return;
+  }
+  remote_addr = inet_addr(IPC_LOCAL_HOST);
+  if(remote_addr == INADDR_NONE) {
+    sock = INVALID_SOCKET;
+    return;
+  }
+#else
   if (sock >= 0) {
     return;
   }
-
+#endif
   if (ipc_make_name () < 0) {
+#ifdef WIN32
+    sock = INVALID_SOCKET;
+#else
     sock = -1;
+#endif
     return;
   }
 
+#ifdef WIN32
+  sock = socket (AF_INET, SOCK_STREAM, 0);
+#ifdef IPC_DEBUG
+  fprintf(stderr, "tex: Socket handle is %d\n", sock);
+  if(sock == INVALID_SOCKET) {
+    fprintf(stderr, "tex: Socket is invalid.\n");
+    return;
+  }
+#endif
+#else
   sock = socket (PF_UNIX, SOCK_STREAM, 0);
+#endif
+
+#ifdef WIN32
+  if (sock != INVALID_SOCKET) {
+#else
   if (sock >= 0) {
+#endif
+#ifdef WIN32
+    sockcntl = 1024;
+    if (connect (sock, (struct sockaddr *)&ipc_addr, ipc_addr_len) == SOCKET_ERROR
+	|| ioctlsocket (sock, O_NONBLOCK, &sockcntl) != 0) {
+      closesocket (sock);
+      sock = INVALID_SOCKET;
+      WSACleanup();
+#ifdef IPC_DEBUG
+    fputs ("tex: IPC socket cannot be connected.\n", stderr);
+    fputs ("tex: Socket is closed.\n", stderr);
+#endif
+      return;
+    }
+#else
     if (connect (sock, ipc_addr, ipc_addr_len) != 0
         || fcntl (sock, F_SETFL, O_NONBLOCK) < 0) {
       close (sock);
       sock = -1;
       return;
     }
+#endif
 #ifdef IPC_DEBUG
     fputs ("tex: Successfully opened IPC socket.\n", stderr);
 #endif
   }
 }
 
-
 static void
 ipc_close_out P1H(void)
 {
@@ -398,8 +563,14 @@ ipc_close_out P1H(void)
   fputs ("tex: Closing output socket ...\n", stderr);
 #endif
   if (ipc_is_open ()) {
+#ifdef WIN32
+    closesocket (sock);
+    sock = INVALID_SOCKET;
+    WSACleanup();
+#else
     close (sock);
     sock = -1;
+#endif
   }
 }
 
@@ -413,13 +584,12 @@ ipc_snd P3C(int, n,  int, is_eof,  char *, data)
     char more_data[1024];
   } ourmsg;
 
-#ifdef IPC_DEBUG
-  fputs ("tex: Sending message to socket ...\n", stderr);
-#endif
   if (!ipc_is_open ()) {
     return;
   }
-
+#ifdef IPC_DEBUG
+  fputs ("tex: Sending message to socket ...\n", stderr);
+#endif
   ourmsg.msg.namelength = n;
   ourmsg.msg.eof = is_eof;
   if (n) {
@@ -427,27 +597,34 @@ ipc_snd P3C(int, n,  int, is_eof,  char *, data)
   }
   n += sizeof (struct msg);
 #ifdef IPC_DEBUG
+  fprintf(stderr, "%d\t%d\n", ourmsg.msg.namelength, ourmsg.msg.eof);
   fputs ("tex: Writing to socket...\n", stderr);
 #endif
+#ifdef WIN32
+  if (send (sock, (char *)&ourmsg, n, 0) != n) {
+    ipc_close_out ();
+  }
+#else
   if (write (sock, &ourmsg, n) != n) {
     ipc_close_out ();
   }
+#endif
 #ifdef IPC_DEBUG
-  fputs ("tex: IPC message sent.\n", stderr);
+  fprintf (stderr, "tex: IPC message sent (%d bytes).\n", n);
 #endif
 }
 
-
 /* This routine notifies the server if there is an eof, or the filename
    if a new DVI file is starting.  This is the routine called by TeX.
-   Omega defines str_start(#) as str_start_ar(# - biggest_char), with
-   biggest_char = 65535 ([4.38], [1.12] om16bit.c).  */
+   Omega defines str_start(#) as str_start_ar[# - too_big_char], with
+   too_big_char = biggest_char + 1 = 65536 (omstr.ch).  */
 
 void
 ipcpage P1C(int, is_eof)
 {
   static boolean begun = false;
   unsigned len = 0;
+  unsigned i;
   string p = "";
 
   if (!begun) {
@@ -458,14 +635,15 @@ ipcpage P1C(int, is_eof)
 #ifndef Omega
     len = strstart[outputfilename + 1] - strstart[outputfilename];
 #else
-    len = strstartar[outputfilename + 1 - 65535L] -
-            strstartar[outputfilename - 65535L];
+    len = strstartar[outputfilename + 1 - 65536L] -
+            strstartar[outputfilename - 65536L];
 #endif
     name = xmalloc (len + 1);
 #ifndef Omega
     strncpy (name, &strpool[strstart[outputfilename]], len);
 #else
-    strncpy (name, &strpool[strstartar[outputfilename - 65535L]], len);
+    for (i=0; i<len; i++)
+      name[i] =  strpool[i+strstartar[outputfilename - 65536L]];
 #endif
     name[len] = 0;
     
@@ -475,6 +653,14 @@ ipcpage P1C(int, is_eof)
     p = concat3 (cwd, DIR_SEP_STRING, name);
     free (name);
     len = strlen(p);
+#ifdef WIN32
+    { char *q;
+      for(q=p; *q; q++) {
+	if((*q == '/') && ( (q == p) || !isknj(*(q-1))))
+	  *q = '\\';
+      }
+    }
+#endif
     begun = true;
   }
   ipc_snd (len, is_eof, p);
@@ -627,41 +813,48 @@ setupcharset P1H(void)
 
 /* SunOS cc can't initialize automatic structs, so make this static.  */
 static struct option long_options[]
-  = { { DUMP_OPTION,		1, 0, 0 },
-      { "help",                 0, 0, 0 },
-      { "ini",			0, (int *) &iniversion, 1 },
-      { "interaction",          1, 0, 0 },
-      { "kpathsea-debug",	1, 0, 0 },
-      { "progname",             1, 0, 0 },
-      { "version",              0, 0, 0 },
+  = { { DUMP_OPTION,              1, 0, 0 },
+      { "help",                   0, 0, 0 },
+      { "ini",                    0, &iniversion, 1 },
+      { "interaction",            1, 0, 0 },
+      { "kpathsea-debug",         1, 0, 0 },
+      { "progname",               1, 0, 0 },
+      { "version",                0, 0, 0 },
+      { "recorder",               0, &recorder_enabled, 1 },
 #ifdef TeX
 #ifdef IPC
-      { "ipc",			0, (int *) &ipcon, 1 },
-      { "ipc-start",		0, (int *) &ipcon, 2 },
+      { "ipc",                    0, &ipcon, 1 },
+      { "ipc-start",              0, &ipcon, 2 },
 #endif /* IPC */
 #ifndef Omega
-      { "mltex",		0, (int *) &mltexp, 1 },
+      { "mltex",                  0, &mltexp, 1 },
 #endif /* !Omega */
-      { "output-comment",	1, 0, 0 },
-      { "shell-escape",		0, (int *) &shellenabledp, 1 },
-      { "debug-format",		0, (int *) &debugformatfile, 1 },
+      { "output-comment",         1, 0, 0 },
+      { "shell-escape",           0, &shellenabledp, 1 },
+      { "debug-format",           0, &debugformatfile, 1 },
+      { "src-specials",           2, 0, 0 },
 #endif /* TeX */
 #if defined (TeX) || defined (MF) || defined (MP)
+      { "file-line-error-style",  0, &filelineerrorstylep, 1 },
+      { "jobname",                1, 0, 0 },
+      { "parse-first-line",       0, &parsefirstlinep, 1 },
 #ifndef Omega
-      { "translate-file",	1, 0, 0 },
+      { "translate-file",         1, 0, 0 },
+      { "default-translate-file", 1, 0, 0 },
 #endif /* !Omega */
 #endif /* TeX || MF || MP */
 #if defined (TeX) || defined (MF)
-      { "mktex",                1, 0, 0 },
-      { "no-mktex",             1, 0, 0 },
+      { "mktex",                  1, 0, 0 },
+      { "no-mktex",               1, 0, 0 },
 #endif /* TeX or MF */
 #ifdef MP
-      { "T",			0, (int *) &troffmode, 1 },
-      { "troff",		0, (int *) &troffmode, 1 },
+      { "T",                      0, &troffmode, 1 },
+      { "troff",                  0, &troffmode, 1 },
+      { "tex",                    1, 0, 0 },
 #endif /* MP */
       { 0, 0, 0, 0 } };
 
-
+#include "jmpost.h"
 static void
 parse_options P2C(int, argc,  string *, argv)
 {
@@ -669,13 +862,13 @@ parse_options P2C(int, argc,  string *, argv)
   int option_index;
 
   for (;;) {
-    g = getopt_long_only (argc, argv, "", long_options, &option_index);
+    g = getopt_long_only (argc, argv, "+", long_options, &option_index);
 
     if (g == -1) /* End of arguments, exit the loop.  */
       break;
 
     if (g == '?') { /* Unknown option.  */
-      usage (1, argv[0]);
+      usage (argv[0]);
     }
 
     assert (g == 0); /* We have no short option names.  */
@@ -686,6 +879,9 @@ parse_options P2C(int, argc,  string *, argv)
     } else if (ARGUMENT_IS ("progname")) {
       user_progname = optarg;
 
+    } else if (ARGUMENT_IS ("jobname")) {
+      job_name = optarg;
+      
     } else if (ARGUMENT_IS (DUMP_OPTION)) {
       dump_name = optarg;
       if (!user_progname) user_progname = optarg;
@@ -704,25 +900,50 @@ parse_options P2C(int, argc,  string *, argv)
         outputcomment[255] = 0;
       }
 
+#ifdef WIN32
+#include <kpathsea/win32lib.h>
+#endif
 #ifdef IPC
     } else if (ARGUMENT_IS ("ipc-start")) {
       ipc_open_out ();
       /* Try to start up the other end if it's not already.  */
       if (!ipc_is_open ()) {
+#ifdef WIN32
+        if (spawnlp (P_NOWAIT, IPC_SERVER_CMD, IPC_SERVER_CMD, NULL) == -1) {
+#else
         if (system (IPC_SERVER_CMD) == 0) {
+#endif
           unsigned i;
+#ifndef WIN32
           for (i = 0; i < 20 && !ipc_is_open (); i++) {
             sleep (2);
+#else
+          for (i = 0; i < 1 && !ipc_is_open (); i++) {
+            Sleep (2);
+#endif
             ipc_open_out ();
           }
-        }
-     }
+	}
+      }
 #endif /* IPC */
+    } else if (ARGUMENT_IS ("src-specials")) {
+       last_source_name = xstrdup("");
+       /* Option `--src" without any value means `auto' mode. */
+       if (optarg == 0) {
+         insertsrcspecialeverypar = true;
+         insertsrcspecialauto = true;
+         srcspecialsoption = true;
+         srcspecialsp = true;
+       } else {
+          parse_src_specials_option(optarg);
+       }
 #endif /* TeX */
 #if defined (TeX) || defined (MF) || defined (MP)
 #ifndef Omega
     } else if (ARGUMENT_IS ("translate-file")) {
       translate_filename = optarg;
+    } else if (ARGUMENT_IS ("default-translate-file")) {
+      default_translate_filename = optarg;
 #endif /* !Omega */
 #endif /* TeX || MF || MP */
 
@@ -733,7 +954,10 @@ parse_options P2C(int, argc,  string *, argv)
     } else if (ARGUMENT_IS ("no-mktex")) {
       kpse_maketex_option (optarg, false);
 #endif /* TeX or MF */
-
+#if defined (MP)
+    } else if (ARGUMENT_IS ("tex")) {
+      mpost_tex_program = optarg;
+#endif /* MP */
     } else if (ARGUMENT_IS ("interaction")) {
         /* These numbers match @d's in *.ch */
       if (STREQ (optarg, "batchmode")) {
@@ -749,12 +973,7 @@ parse_options P2C(int, argc,  string *, argv)
       }
       
     } else if (ARGUMENT_IS ("help")) {
-      string help = PROGRAM_HELP;
-#if defined (TeX) && defined (IPC)
-      /* Don't say the options exist unless they really do.  */
-      help = concat (help, TEX_IPC_HELP);
-#endif
-      usage (0, help);
+       usagehelp (PROGRAM_HELP);
 
     } else if (ARGUMENT_IS ("version")) {
       printversionandexit (BANNER, COPYRIGHT_HOLDER, AUTHOR);
@@ -763,6 +982,57 @@ parse_options P2C(int, argc,  string *, argv)
   }
 }
 
+#if defined(TeX)
+void 
+parse_src_specials_option P1C(const_string, opt_list)
+{
+  char * toklist = xstrdup(opt_list);
+  char * tok;
+  insertsrcspecialauto = false;
+  tok = strtok (toklist, ", ");
+  while (tok) {
+    if (strcmp (tok, "everypar") == 0
+        || strcmp (tok, "par") == 0
+        || strcmp (tok, "auto") == 0) {
+      insertsrcspecialauto = true;
+      insertsrcspecialeverypar = true;
+    } else if (strcmp (tok, "everyparend") == 0
+               || strcmp (tok, "parend") == 0)
+      insertsrcspecialeveryparend = true;
+    else if (strcmp (tok, "everycr") == 0
+             || strcmp (tok, "cr") == 0)
+      insertsrcspecialeverycr = true;
+    else if (strcmp (tok, "everymath") == 0
+             || strcmp (tok, "math") == 0)
+      insertsrcspecialeverymath = true;
+    else if (strcmp (tok, "everyhbox") == 0
+             || strcmp (tok, "hbox") == 0)
+      insertsrcspecialeveryhbox = true;
+    else if (strcmp (tok, "everyvbox") == 0
+             || strcmp (tok, "vbox") == 0)
+      insertsrcspecialeveryvbox = true;
+    else if (strcmp (tok, "everydisplay") == 0
+             || strcmp (tok, "display") == 0)
+      insertsrcspecialeverydisplay = true;
+    else if (strcmp (tok, "none") == 0) {
+      /* This one allows to reset an option that could appear in texmf.cnf */
+      insertsrcspecialauto = insertsrcspecialeverypar = 
+        insertsrcspecialeveryparend = insertsrcspecialeverycr = 
+        insertsrcspecialeverymath =  insertsrcspecialeveryhbox =
+        insertsrcspecialeveryvbox = insertsrcspecialeverydisplay = false;
+    } else {
+      WARNING1 ("Ignoring unknown argument `%s' to --src-specials", tok);
+    }
+    tok = strtok(0, ", ");
+  }
+  free(toklist);
+  srcspecialsp=insertsrcspecialauto | insertsrcspecialeverypar |
+    insertsrcspecialeveryparend | insertsrcspecialeverycr |
+    insertsrcspecialeverymath |  insertsrcspecialeveryhbox |
+    insertsrcspecialeveryvbox | insertsrcspecialeverydisplay;
+  srcspecialsoption = true;
+}
+#endif
 
 /* If the first thing on the command line (we use the globals `argv' and
    `optind') is a normal filename (i.e., does not start with `&' or
@@ -770,22 +1040,12 @@ parse_options P2C(int, argc,  string *, argv)
    FORMAT is a readable dump file, then set DUMP_VAR to FORMAT.
    Also call kpse_reset_program_name to ensure the correct paths for the
    format are used.  */
-
 static void
-maybe_parse_first_line P1H(void)
+parse_first_line P1H(void)
 {
   char first_char = optind < argc ? argv[optind][0] : 0;
-  string parse_first_line = kpse_var_value ("parse_first_line");
 
-  /* Default to parsing the first line. */
-  if (!parse_first_line) parse_first_line = "t";
-
-  if ((*parse_first_line == 't'
-       || *parse_first_line == 'y'
-       || *parse_first_line == '1')
-      && first_char
-      && first_char != '&'
-      && first_char != '\\') {
+  if (first_char && first_char != '&' && first_char != '\\') {
     /* If the file can't be found, don't look too hard now.  We'll
        detect that it's missing in the normal course of things and give
        the error then.  */
@@ -796,6 +1056,8 @@ maybe_parse_first_line P1H(void)
       xfclose (f, in_name);
 
       /* We deal with the general format "%&fmt --translate-file=tcx" */
+      /* The idea of using this format came from Wlodzimierz Bzyl
+         <matwb@monika.univ.gda.pl> */
       if (first_line && first_line[0] == '%' && first_line[1] == '&') {
         /* Parse the first line into at most three space-separated parts. */
         char *s;
@@ -846,6 +1108,7 @@ maybe_parse_first_line P1H(void)
             s = *parse+16;
           }
           /* Just set the name, no sanity checks here. */
+          /* FIXME: remove trailing spaces. */
           if (s && *s) {
             translate_filename = xstrdup(s);
           }
@@ -888,8 +1151,11 @@ opennameok P3C(const_string, fname, const_string, check_var,
     const_string base = xbasename (fname);
     /* Disallow .rhosts, .login, etc.  Allow .tex (for LaTeX).  */
     if (base[0] == 0 ||
-        (base[0] == '.' && !IS_DIR_SEP(base[1]) && !STREQ (base, ".tex")))
+        (base[0] == '.' && !IS_DIR_SEP(base[1]) && !STREQ (base, ".tex"))) {
+      fprintf(stderr, "%s: Not writing to %s (%s = %s).\n",
+              program_invocation_name, fname, check_var, open_choice);
       return false;
+    }
   }
 #else
   /* Other OSs don't have special names? */
@@ -905,12 +1171,17 @@ opennameok P3C(const_string, fname, const_string, check_var,
        fname begins the TEXMFOUTPUT, and is followed by / */
     if (!texmfoutput || *texmfoutput == '\0'
         || fname != strstr (fname, texmfoutput)
-        || !IS_DIR_SEP(fname[strlen(texmfoutput)]))
+        || !IS_DIR_SEP(fname[strlen(texmfoutput)])) {
+      fprintf(stderr, "%s: Not writing to %s (%s = %s).\n",
+              program_invocation_name, fname, check_var, open_choice);
       return false;
+    }
   }
   /* For all pathnames, we disallow "../" at the beginning or "/../"
      anywhere.  */
   if (fname[0] == '.' && fname[1] == '.' && IS_DIR_SEP(fname[2])) {
+    fprintf(stderr, "%s: Not writing to %s (%s = %s).\n",
+            program_invocation_name, fname, check_var, open_choice);
     return false;
   } else {
     const_string dotpair = strstr (fname, "..");
@@ -918,11 +1189,10 @@ opennameok P3C(const_string, fname, const_string, check_var,
     if (dotpair && IS_DIR_SEP(dotpair[2]) && IS_DIR_SEP(dotpair[-1]))
       return false;
   }
-
-  /* We passed all tests.  */
   return true;
 }
 
+
 boolean openinnameok P1C(const_string, fname)
 {
     /* For input default to all. */
@@ -1020,9 +1290,19 @@ get_date_and_time P4C(integer *, minutes,  integer *, day,
 integer
 getrandomseed()
 {
+#if defined (HAVE_GETTIMEOFDAY)
+  struct timeval tv;
+  gettimeofday(&tv, NULL);
+  return (tv.tv_usec + 1000000 * tv.tv_usec);
+#elif defined (HAVE_FTIME)
+  struct timeb tb;
+  ftime(&tb);
+  return (tb.millitm + 1000 * tb.time);
+#else
   time_t clock = time ((time_t*)NULL);
   struct tm *tmptr = localtime(&clock);
   return (tmptr->tm_sec + 60*(tmptr->tm_min + 60*tmptr->tm_hour));
+#endif
 }
 
 /* Read a line of input as efficiently as possible while still looking
@@ -1068,11 +1348,10 @@ input_line P1C(FILE *, f)
         buffer[last++] = i;
       } else {
         /* JIS encoding */
-#ifdef SJISPTEX
-        i = JIStoSJIS(i << 8 | getc(f));
-#else
-        i = JIStoEUC(i << 8 | getc(f));
-#endif
+	if(prockanjicode == 2)
+	  i = JIStoSJIS(i << 8 | getc(f));
+	else if(prockanjicode == 1)
+	  i = JIStoEUC(i << 8 | getc(f));
         buffer[last++] = (i >> 8) & 0xff;
         buffer[last++] = i & 0xff;
       }
@@ -1083,7 +1362,7 @@ input_line P1C(FILE *, f)
     buffer[last++] = i;
 #endif /* KANJI */
 
-  if (i == EOF && last == first)
+  if (i == EOF && errno != EINTR && last == first)
     return false;
 
   /* We didn't get the whole line because our buffer was too small.  */
@@ -1100,7 +1379,8 @@ input_line P1C(FILE *, f)
 
   /* If next char is LF of a CRLF, read it.  */
   if (i == '\r') {
-    i = getc (f);
+    while ((i = getc (f)) == EOF && errno == EINTR)
+      ;
     if (i != '\n')
       ungetc (i, f);
   }
@@ -1128,7 +1408,7 @@ static char *edit_value = EDITOR;
    actual filename starts; FNLENGTH is how long the filename is.  */
    
 void
-calledit P4C(ASCIIcode *, filename,
+calledit P4C(packedASCIIcode *, filename,
              poolpointer, fnstart,
              integer, fnlength,
              integer, linenumber)
@@ -1199,7 +1479,11 @@ calledit P4C(ASCIIcode *, filename,
   *temp = 0;
 
   /* Execute the command.  */
+#ifdef WIN32
+  if (win32_system (command) != 0)
+#else
   if (system (command) != 0)
+#endif
     fprintf (stderr, "! Trouble executing `%s'.\n", command);
 
   /* Quit, since we found an error.  */
@@ -1342,7 +1626,7 @@ setupboundvariable P3C(integer *, var,  const_string, var_name,  integer, dflt)
     integer conf_val = atoi (expansion);
     /* It's ok if the cnf file specifies 0 for extra_mem_{top,bot}, etc.
        But negative numbers are always wrong.  */
-    if (conf_val < 0 || conf_val == 0 && dflt > 0) {
+    if (conf_val < 0 || (conf_val == 0 && dflt > 0)) {
       fprintf (stderr,
                "%s: Bad value (%ld) in texmf.cnf for %s, keeping %ld.\n",
                program_invocation_name,
@@ -1353,6 +1637,142 @@ setupboundvariable P3C(integer *, var,  const_string, var_name,  integer, dflt)
     free (expansion);
   }
 }
+
+#if defined(TeX) || defined(MP) || defined(MF)
+static void
+checkpoolpointer (poolpointer poolptr, size_t len)
+{
+  if (poolptr + len >= poolsize) {
+    fprintf (stderr, "\nstring pool overflow [%i bytes]\n", 
+	     (int)poolsize); /* fixme */
+    exit(1);
+  }
+}
+
+#if !defined(pdfTeX) && !defined(pdfeTeX)
+static int
+maketexstring(const_string s)
+{
+  size_t len;
+  assert (s != 0);
+  len = strlen(s);
+  checkpoolpointer (poolptr, len);
+  while (len-- > 0)
+    strpool[poolptr++] = *s++;
+  return (makestring());
+}
+#endif
+
+strnumber
+makefullnamestring()
+{
+  return maketexstring(fullnameoffile);
+}
+
+strnumber
+getjobname()
+{
+    strnumber ret = curname;
+    if (job_name != NULL)
+      ret = maketexstring(job_name);
+    return ret;
+}
+#endif
+
+#if defined(TeX)
+int
+compare_paths P2C(const_string, p1, const_string, p2)
+{
+  int ret;
+  while (
+#ifdef MONOCASE_FILENAMES
+		 (((ret = (toupper(*p1) - toupper(*p2))) == 0) && (*p2 != 0))
+#else
+         (((ret = (*p1 - *p2)) == 0) && (*p2 != 0))
+#endif
+		 || (IS_DIR_SEP(*p1) && IS_DIR_SEP(*p2))) {
+	p1++, p2++;
+  }
+  ret = (ret < 0 ? -1 : (ret > 0 ? 1 : 0));
+  return ret;
+}
+
+string
+gettexstring P1C(strnumber, s)
+{
+  size_t i, len;
+  string name;
+
+#ifndef Omega
+  len = strstart[s + 1] - strstart[s];
+#else
+  len = strstartar[s + 1 - 65536L] -
+	strstartar[s - 65536L];
+#endif
+  name = xmalloc (len + 1);
+#ifndef Omega
+  strncpy (name, &strpool[strstart[s]], len);
+#else
+  /* Don't use strncpy.  The strpool is not made up of chars. */
+  for (i=0; i<len; i++) name[i] =  strpool[i+strstartar[s - 65536L]];
+#endif
+  name[len] = 0;
+  return name;
+}
+
+boolean
+isnewsource P2C(strnumber, srcfilename, int, lineno)
+{
+  size_t len;
+  char *name = gettexstring(srcfilename);
+  return (compare_paths(name, last_source_name) != 0 || lineno != last_lineno);
+}
+
+void
+remembersourceinfo P2C(strnumber, srcfilename,
+					   int, lineno)
+{
+  size_t len;
+
+  if (last_source_name)
+	free(last_source_name);
+  last_source_name = gettexstring(srcfilename);
+  last_lineno = lineno;
+}
+
+poolpointer
+makesrcspecial P2C(strnumber, srcfilename,
+				   int,	lineno)
+{
+  char fname[128];
+  poolpointer oldpoolptr = poolptr;
+  char *filename = gettexstring(srcfilename);
+  char buf[128];
+  char * s = buf;  
+
+  if((*filename == '.') && IS_DIR_SEP(filename[1]))
+    strcpy(fname, filename + 2);
+  else
+    strcpy(fname, filename);
+
+  sprintf (buf, "src:%d%s", lineno, 
+		   (isdigit(*fname) ? " " : ""));
+
+  if (poolptr + strlen(buf) + strlen(fname) >= poolsize) {
+	fprintf (stderr, "\nstring pool overflow\n"); /* fixme */
+	exit (1);
+  }
+  s = buf;
+  while (*s)
+    strpool[poolptr++] = *s++;
+
+  s = fname;
+  while (*s)
+    strpool[poolptr++] = *s++;
+	
+  return (oldpoolptr);
+}
+#endif
 
 #ifdef MP
 /* Invoke makempx (or troffmpx) to make sure there is an up-to-date
@@ -1380,8 +1800,15 @@ callmakempx P2C(string, mpname,  string, mpxname)
     if (!cnf_cmd)
       cnf_cmd = xstrdup (MPXCOMMAND);
 
-    cmd = concatn (cnf_cmd, troffmode ? " -troff " : " ",
-                   mpname, " ", mpxname, NULL);
+    if (troffmode)
+      cmd = concatn (cnf_cmd, " -troff ",
+		     mpname, " ", mpxname, NULL);
+    else if (mpost_tex_program && *mpost_tex_program)
+      cmd = concatn (cnf_cmd, " -tex=", mpost_tex_program, " ",
+		     mpname, " ", mpxname, NULL);
+    else
+      cmd = concatn (cnf_cmd, " -tex ", mpname, " ", mpxname, NULL);
+
     /* Run it.  */
     ret = system (cmd);
     free (cmd);
@@ -1756,6 +2183,9 @@ boolean
 initscreen P1H(void)
 {
   /* If MFTERM is set, use it.  */
+#ifdef WIN32
+  int retval;
+#endif
   char *tty_type = kpse_var_value ("MFTERM");
   
   if (tty_type == NULL)
@@ -1785,8 +2215,11 @@ initscreen P1H(void)
   for (mfwp = mfwsw; mfwp->mfwsw_type != NULL; mfwp++) {
     if (!strncmp (mfwp->mfwsw_type, tty_type, strlen (mfwp->mfwsw_type))
 	|| STREQ (tty_type, "emacs"))
-      if (mfwp->mfwsw_initscreen)
-	return ((*mfwp->mfwsw_initscreen) ());
+      if (mfwp->mfwsw_initscreen) {
+	retval = ((*mfwp->mfwsw_initscreen) ());
+	Sleep(1000); /* Wait for opening a window --ak */
+	return retval;
+      }
       else {
         fprintf (stderr, "mf: Couldn't initialize online display for `%s'.\n",
                  tty_type);
